package rione.agent;

import java.util.*;

import rescuecore2.messages.Command;
import rescuecore2.standard.components.*;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import rione.model.AdvancedWorldModel;

public abstract class Agent<E extends StandardEntity> extends StandardAgent<E> {

	/** WorldModelへの参照 */
	protected AdvancedWorldModel world;

	/** 現在の時間 */
	protected int time;

	protected ChangeSet changed;

	/** 各サイクルに送信したコマンドの履歴 <時間, コマンド> */
	private Map<Integer, StandardMessageURN> commandHistory = new HashMap<Integer, StandardMessageURN>();

	/** シミュレーションを行うサイクル数 */
	protected int timestep;

	/** エージェントの思考可能時間 */
	protected int thinkTime;

	@Override
	protected void postConnect() {
		super.postConnect();
		initialize();
	}

	/**
	 * エージェントの初期化を行う。 必要な時にオーバーライドして使う。
	 * オーバーライドする際は、メソッド内で必ずスーパークラスのinitialize()を呼び出すこと。
	 */
	protected void initialize() {

		model.indexClass(StandardEntityURN.AMBULANCE_TEAM, StandardEntityURN.FIRE_BRIGADE, StandardEntityURN.POLICE_FORCE, StandardEntityURN.BUILDING, StandardEntityURN.AMBULANCE_CENTRE, StandardEntityURN.FIRE_STATION, StandardEntityURN.POLICE_OFFICE, StandardEntityURN.ROAD, StandardEntityURN.REFUGE, StandardEntityURN.BUILDING, StandardEntityURN.BLOCKADE, StandardEntityURN.CIVILIAN);
		world = (AdvancedWorldModel) model;
		world.initialize(me(), config);
		timestep = world.getConfig().timestep;
		thinkTime = world.getConfig().thinkTime;
		random = new Random(getID().getValue());
	}

	@Override
	protected StandardWorldModel createWorldModel() {
		return new AdvancedWorldModel();
	}

	protected abstract void act() throws ActionCommandException;

	/**
	 * actの前に呼び出すメソッド。 必要な時にオーバーライドして使う。 メソッド内でスーパークラスのこのメソッドを呼び出すこと。
	 */
	protected void prepareForAct() {
	}

	/**
	 * コマンドを送信した後に呼び出すメソッド。 必要な時にオーバーライドして使う。 メソッド内でスーパークラスのこのメソッドを呼び出すこと。
	 */
	protected void afterAct() {
	}

	/**
	 * 受け取った情報を処理する
	 * 
	 * @param heard
	 */
	protected abstract void hear(Collection<Command> heard);

	@Override
	protected void think(int time, ChangeSet changed, Collection<Command> heard) {
		try { // 試合時にはこのコメントアウトを外す
			this.time = time;
			world.setTime(time);
			this.changed = changed;
			hear(heard);
			prepareForAct();
			try {
				if (world.getConfig().ignoreUntil <= time) {
					act();
				}
				rest();
			}
			catch (ActionCommandException ace) {
				commandHistory.put(time, ace.message);
			}
			afterAct();
			world.setTime(world.getTime() + 1); // thinkメソッドが呼ばれる前にWorldModelのtimeが参照される可能性があるので、1足しておく
		}
		catch (Exception e) { // 試合時にはこのコメントアウトを外す
			System.err.println("Exception : time = " + time + " id = " + me().getID());
			e.printStackTrace();
		}
		catch (Error e) {
			System.err.println("Error : time = " + time + " id = " + me().getID());
			e.printStackTrace();
		}
	}

	protected void rest() throws ActionCommandException {
		sendRest(time);
		throw new ActionCommandException(StandardMessageURN.AK_REST);
	}

	protected StandardMessageURN getCommandHistory(int time) {
		return commandHistory.get(time);
	}

	
	/**
	 * 自身の背番号を返す。
	 * 
	 * @return 背番号(0~)
	 */
	protected int getUniform() {
		return world.getUniform().toUniform(getID());
	}

	/**
	 * Entityが視界内にあるかどうかを返す
	 * 
	 * @param id
	 *            判定するEntityのID
	 * @return 視界内にあるときtrue，ないときfalse
	 */
	protected boolean isVisible(EntityID id) {
		return changed.getChangedEntities().contains(id);
	}

	/**
	 * Entityが視界内にあるかどうかを返す
	 * 
	 * @param id
	 *            判定するEntity
	 * @return 視界内にあるときtrue，ないときfalse
	 */
	protected boolean isVisible(StandardEntity entity) {
		return isVisible(entity.getID());
	}

	/**
	 * 視界内にあるEntityの集合を返す． 自分も含める．
	 * 
	 * @return 視界内のEntityの集合
	 */
	public Set<EntityID> getVisibleEntities() {
		if (changed != null) {
			return changed.getChangedEntities();
		}
		return null;
	}

	@SuppressWarnings("serial")
	public static class ActionCommandException extends Exception {
		private StandardMessageURN message;

		public ActionCommandException(StandardMessageURN message) {
			super();
			this.message = message;
		}
	}
}
