package rione.agent;

import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.messages.AKMove;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.EntityID;
import rione.model.route.pov.CostFunction;
import rione.model.route.pov.POVRouter;

public abstract class HumanoidAgent<E extends Human> extends CommunicationAgent<E> {

	protected POVRouter router;

	@Override
	protected void initialize() {
		super.initialize();
		router = world.getRouter();
	}

	Pair<Integer, Integer> preLocation = null;
	private double distAverage = 0.0;
	private int distCount = 0;

	@Override
	protected void prepareForAct() {
		world.update(me(), changed);
		super.prepareForAct();
		router.update(me().getPosition(), getVisibleEntities());
		updateMoveAverage();
	}

	/**
	 * 移動距離の平均を更新
	 */
	private void updateMoveAverage() {
		final Pair<Integer, Integer> location = me().getLocation(world);
		if(getCommandHistory(time-1) != null){
			if (getCommandHistory(world.getTime() - 1) == StandardMessageURN.AK_MOVE) {
				if (preLocation != null) {
					final double dist = Math.hypot(location.first() - preLocation.first(),
							location.second() - preLocation.second());
					distCount++;
					if (distAverage == 0.0) {
						distAverage = dist;
					}
					else {
						// distAverage * distCount すると浮動小数点演算バグになりそうなのでこっちを先に計算する．
						double t = (double) (distCount - 1) / distCount;
						distAverage = distAverage * t + dist / distCount; 
					}
				}
			}
		}
		preLocation = location;
	}

	protected void moveFront(StandardEntity destination, CostFunction costFunc) throws ActionCommandException {
		if (destination instanceof Human) {
			Human human = (Human) destination;
			StandardEntity position = human.getPosition(world);
			if (position instanceof Area) {
				moveFront(router.getAStar(me(), (Area) position, costFunc), destination.getLocation(world));
			}
			else {
				// なんらかの理由でエージェントがATの上にある場合，
				// もしくは現在位置がnullの場合エラーになる
				if (position == null) {
					System.err.println("Move error\ttime=" + time + " id=" + me().getID() + " my position is null");
				}
				else {
					System.err.println("Move error\ttime=" + time + " id=" + me().getID() + " move from " + position);
				}
				return;
			}
		}
		else if (destination instanceof Blockade) {
			Blockade block = (Blockade) destination;
			EntityID position = block.getPosition();
			moveFront(router.getAStar(me(), (Area) world.getEntity(position), costFunc), destination.getLocation(world));
		}
		moveFront(router.getAStar(me(), (Area) destination, costFunc), destination.getLocation(world));
	}

	private void moveFront(List<EntityID> path, Pair<Integer, Integer> location) throws ActionCommandException {
		if (location == null) {
			moveFront(path, -1, -1);
		}
		else {
			moveFront(path, location.first(), location.second());
		}
	}

	/**
	 * パスと目標地点の座標を渡したmove
	 * 
	 * @param path
	 *            目標地点までのパス
	 * @param x
	 *            目標地点のX座標
	 * @throws ActionCommandException
	 */
	protected void move(List<EntityID> path, int x, int y) throws ActionCommandException {
		send(new AKMove(getID(), time, path, x, y));
		throw new ActionCommandException(StandardMessageURN.AK_MOVE);
	}
	
	/**
	 * 集合に対するmove<br>
	 * Areaの集合を渡さなくてはいけませんが，
	 * world.getEntityOfTypeなどの観点からStandardEntityの集合で定義しています．
	 * 
	 * @param destinations
	 * @throws ActionCommandException
	 */

	protected void moveFront(Collection<? extends StandardEntity> destinations) throws ActionCommandException {
		moveFront(destinations, router.getNormalCostFunction());
	}

	/**
	 * 集合に対するmStandardEntityove<br>
	 * Areaの集合を渡さなくてはいけませんが，
	 * world.getEntityOfTypeなどの観点からStandardEntityの集合で定義しています．
	 * 
	 * @param destinations
	 * @param costFunc
	 * @throws ActionCommandException
	 */

	protected void moveFront(Collection<? extends StandardEntity> destinations, CostFunction costFunc) throws ActionCommandException {
		moveFront(router.getMultiDest(location(), destinations, costFunc));
	}

	/**
	 * パスを渡したmove
	 * 
	 * @param path
	 *            目標地点までのパス
	 * @throws ActionCommandException
	 */

	protected void moveFront(List<EntityID> path) throws ActionCommandException {
		moveFront(path, -1, -1);
	}

	protected void moveFront(List<EntityID> path, int x, int y) throws ActionCommandException {
		int last = path.size() - 1;
		if (world.getEntity(path.get(last)) instanceof Building) {
			path.remove(last);
		}
		send(new AKMove(getID(), time, path, x, y));
		throw new ActionCommandException(StandardMessageURN.AK_MOVE);
	}

	/**
	 * 目的AreaとXY座標を指定したmove
	 * 
	 * @param destination
	 *            目的Area
	 * @param x
	 *            目的X座標
	 * @param y
	 *            目的Y座標
	 * @throws ActionCommandException
	 *             この移動方法を指定したい場合は，目標地点をArea型で指定する．
	 */
	protected void move(Area destination, int x, int y) throws ActionCommandException {
		move(router.getAStar(me(), destination, router.getNormalCostFunction()), x, y);
	}

	/**
	 * パスと目標地点の座標を渡したmove
	 * 
	 * @param path
	 *            目標地点までのパス
	 * @param location
	 *            目標地点のX,Y座標
	 * @throws ActionCommandException
	 */
	protected void move(List<EntityID> path, Pair<Integer, Integer> location) throws ActionCommandException {
		if (location == null) {
			move(path, -1, -1);
		}
		else {
			move(path, location.first(), location.second());
		}
	}

	/**
	 * パスを渡したmove
	 * 
	 * @param path
	 *            目標地点までのパス
	 * @throws ActionCommandException
	 */
	protected void move(List<EntityID> path) throws ActionCommandException {
		move(path, -1, -1);
	}

	/**
	 * 目標エンティティとコストを渡したmove
	 * 
	 * @param destination
	 * @param costFunc
	 * @throws ActionCommandException
	 */
	protected void move(StandardEntity destination, CostFunction costFunc) throws ActionCommandException {
		if (destination instanceof Human) {
			Human human = (Human) destination;
			StandardEntity position = human.getPosition(world);
			if (position instanceof Area) {
				move(router.getAStar(me(), (Area) position, costFunc), destination.getLocation(world));
			}
			else {
				// なんらかの理由でエージェントがATの上にある場合，
				// もしくは現在位置がnullの場合エラーになる
				if (position == null) {
					System.err.println("Move error\ttime=" + time + " id=" + me().getID() + " my position is null");
				}
				else {
					System.err.println("Move error\ttime=" + time + " id=" + me().getID() + " move from " + position);
				}
				return;
			}
		}
		else if (destination instanceof Blockade) {
			Blockade block = (Blockade) destination;
			EntityID position = block.getPosition();
			move(router.getAStar(me(), (Area) world.getEntity(position), costFunc), destination.getLocation(world));
		}
		move(router.getAStar(me(), (Area) destination, costFunc), destination.getLocation(world));
	}

	/**
	 * 目標エンティティを渡したmove
	 * 
	 * @param destination
	 * @throws ActionCommandException
	 */
	protected void move(StandardEntity destination) throws ActionCommandException {
		move(destination, router.getNormalCostFunction());
	}

	/**
	 * 目標のEntityIDとコストを渡したmove
	 * 
	 * @param destination
	 * @param costFunc
	 * @throws ActionCommandException
	 */
	protected void move(EntityID destination, CostFunction costFunc) throws ActionCommandException {
		move(world.getEntity(destination), costFunc);
	}

	/**
	 * 目標のEntityIDを渡したmove
	 * 
	 * @param destination
	 * @throws ActionCommandException
	 */
	protected void move(EntityID destination) throws ActionCommandException {
		move(world.getEntity(destination));
	}

	/**
	 * 足立法用移動アルゴリズム
	 * @param end
	 * 目的地
	 * @param agents
	 * 現在地
	 * @throws ActionCommandException
	 */
	protected void move(Area end,Area agents) throws ActionCommandException{
			move(router.getAdachi(end, agents, router.getSearchCostFunction()));
	}
	/**
	 * 集合に対するmStandardEntityove<br>
	 * Areaの集合を渡さなくてはいけませんが，
	 * world.getEntityOfTypeなどの観点からStandardEntityの集合で定義しています．
	 * 
	 * @param destinations
	 * @param costFunc
	 * @throws ActionCommandException
	 */
	protected void move(Collection<? extends StandardEntity> destinations, CostFunction costFunc) throws ActionCommandException {
		move(router.getMultiDest(location(), destinations, costFunc));
	}

	/**
	 * 集合に対するmove<br>
	 * Areaの集合を渡さなくてはいけませんが，
	 * world.getEntityOfTypeなどの観点からStandardEntityの集合で定義しています．
	 * 
	 * @param destinations
	 * @throws ActionCommandException
	 */
	protected void move(Collection<? extends StandardEntity> destinations) throws ActionCommandException {
		move(destinations, router.getNormalCostFunction());
	}

	/**
	 * 集合に対するmove<br>
	 * Areaの集合を渡さなくてはいけませんが，
	 * world.getEntityOfTypeなどの観点からStandardEntityの集合で定義しています．
	 * 
	 * @param destinations
	 * @param Front
	 *            frontに行きたいならtrue
	 * @throws ActionCommandException
	 */
	protected void moveReachable(Collection<? extends StandardEntity> destinations, boolean front) throws ActionCommandException {
		moveReachable(destinations, router.getNormalCostFunction(), front);
	}

	/**
	 * 集合に対するmStandardEntityove<br>
	 * Areaの集合を渡さなくてはいけませんが，
	 * world.getEntityOfTypeなどの観点からStandardEntityの集合で定義しています．
	 * Reachableな目的地が一つもない場合は，何もしません．
	 * 
	 * @param destinations
	 * @param costFunc
	 * @throws ActionCommandException
	 */
	protected void moveReachable(Collection<? extends StandardEntity> destinations, CostFunction costFunc, boolean front) throws ActionCommandException {
		for (Iterator<? extends StandardEntity> it = destinations.iterator(); it.hasNext();) {
			if (!router.isReachable((Area) it.next())) {
				it.remove();
			}
		}
		if (destinations.isEmpty())
			return;
		moveReachable(router.getMultiDest(location(), destinations, costFunc), front);
	}

	/**
	 * パスを渡したmove
	 * 
	 * @param path
	 *            目標地点までのパス
	 * @throws ActionCommandException
	 */
	protected void moveReachable(List<EntityID> path, boolean front) throws ActionCommandException {
		moveReachable(path, -1, -1, front);
	}

	protected void moveReachable(List<EntityID> path, int x, int y, boolean front) throws ActionCommandException {
		if (front && world.getEntity(path.get(path.size() - 1)) instanceof Building) {
			path.remove(path.size() - 1);
		}
		send(new AKMove(getID(), time, path, x, y));
		throw new ActionCommandException(StandardMessageURN.AK_MOVE);
	}

	/**
	 * 利便性のためAreaにしておく．これに慣れると市民のpositionがCivilianになることを忘れるかも．
	 */
	@Override
	public Area location() {
		return (Area) super.location();
	}

	/**
	 * これまでの移動の平均距離を返します．
	 * @return 
	 */
	public double getDistAverage() {
		// Blockadeにぶつかったりしてまともなデータにならない場合もあるので
		return Math.max(10000.0, distAverage);
	}
}
