package rione.communication;

import java.util.ArrayList;
import java.util.Collection;
import java.util.PriorityQueue;

import rescuecore2.messages.Command;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.ChangeSet;
import rione.communication.channel.ChannelManager;
import rione.io.BitArrayInputStream;
import rione.io.BitArrayOutputStream;
import rione.model.AdvancedWorldModel;
import rione.util.BitUtil;

/**
 * CommunicationのFacade<br>
 * <br>
 * このクラスは通信を管理するクラスです．
 * 具体的な処理を行うCommunicationEventListenerの呼び出しを中心に行っています．
 * @author utisam
 * @see Port
 */
public class CommunicationManager {

	final private AdvancedWorldModel world;

	final private ChannelManager chManager;

	/** 
	 * 無線用<br>
	 * インデックスはポート番号に相当する
	 */
	final private Port[] radioEventListener;
	final private int radioHeaderBitSize;
	/** 
	 * 音声用<br>
	 * インデックスはポート番号に相当する
	 */
	final private Port[] voiceEventListener;
	final private int voiceHeaderBitSize;
	final private CivilianVoiceListener[] civilianListener;

	public CommunicationManager(CommunicationManagerPartsBuilder parts) {
		parts.build();
		
		world = parts.getWorld();
		chManager = parts.getChannelManager();

		radioEventListener = parts.getRadioEventListener();
		voiceEventListener = parts.getVoiceEventListener();
		radioHeaderBitSize = BitUtil.needBitSize(radioEventListener.length);
		voiceHeaderBitSize = BitUtil.needBitSize(voiceEventListener.length);
		civilianListener = parts.getCivilianVoiceListener();
	}

	public void update() {
		chManager.update();
	}

	public int getSendChannel() {
		return chManager.getSendChannel();
	}

	public int getSpeakChannel() {
		return chManager.getSpeakChannel();
	}

	public int[] getSubscribeChannels() {
		return chManager.getSubscribeChannels();
	}

	public byte[] createSendMessage(final int channel, final ChangeSet changed) {
		// test
		// return "test message".getBytes();
		final int limit = world.getConfig().radioChannels.get(channel).bandwidth;
		return createMessage(radioEventListener, radioHeaderBitSize, limit, changed);
	}

	public byte[] createSpeakMessage(final int channel, final ChangeSet changed) {
		// test
		// return "test voice".getBytes();
		final int limit = world.getConfig().voiceChannels.get(channel).size;
		return createMessage(voiceEventListener, voiceHeaderBitSize, limit, changed);
	}

	static private byte[] createMessage(final Port[] listener, final int headerSize, final int limit, final ChangeSet changed) {
		final PriorityQueue<MessageBitSection> que = new PriorityQueue<MessageBitSection>();
		final BitArrayOutputStream stream = new BitArrayOutputStream();
		// 各リスナーのメッセージをヘッダ付きでqueへ
		for (int i = 0; i < listener.length; i++) {
			for (listener[i].init(changed); listener[i].hasNext();) {
				MessageBitSection sec = listener[i].next();
				// TODO staticなハフマン符号化
				sec.setHeaderNumber(i);
				que.add(sec);
			}
		}
		while (!que.isEmpty()) {
			final MessageBitSection sec = que.poll();

			// データサイズの判定
			if ((stream.getBitLength() + sec.getBitLength()) / 8 + 1 > limit) {
				continue;
			}
			// 書きこむ
			stream.writeBit(sec.getHeaderNumber(), headerSize);
			for (Pair<Integer, Integer> writePair : sec) {
				stream.writeBit(writePair.first(), writePair.second());
			}
		}
		return stream.getData();
	}

	public void read(final Collection<Command> heard) {
		for (Command command : heard) {
			if (command instanceof AKSpeak) {
				final AKSpeak message = (AKSpeak) command;
				final StandardEntity sender = world.getEntity(message.getAgentID());
				final int channel = message.getChannel();
				//final byte[] context = message.getContent();
				// System.out.println("sender: " + sender);
				// System.out.println("size: " + context.length);
				// System.out.println("Channel: "
				// + ((chManager.isVoiceChannel(channel)) ? "Voice" : "Radio")
				// + channel);
				if (sender instanceof Civilian) {
					for (CivilianVoiceListener l : civilianListener) {
						l.hear(message);
					}
				}
				else if (chManager.isVoiceChannel(channel)) {
					readMessage(message, voiceEventListener, voiceHeaderBitSize);
				}
				else /* if (chManager.isRadioChannel(channel)) */{
					readMessage(message, radioEventListener, radioHeaderBitSize);
				}
			}
		}
	}

	static private void readMessage(final AKSpeak message, final Port[] listener, final int headerSize) {
		final BitArrayInputStream stream = new BitArrayInputStream(message.getContent());
		// Release版ではportsを削除する
		final ArrayList<String> ports = new ArrayList<String>();
		while (stream.hasNext()) {
			final int header = stream.readBit(headerSize);
			try {
				ports.add(listener[header].getClass().getName());
				listener[header].read(message.getAgentID(), message.getTime(), stream);
			}
			catch (ArrayIndexOutOfBoundsException e) {
				break;
			}
			catch (Exception e) {
				System.out.println(ports);
				e.printStackTrace();
				//System.exit(0);
			}
		}
	}
}
