package rione.dynamictool;

import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.WindowConstants;

/**
 * 実行中にダイアログから動的に定数を設定する
 * @author utisam
 *
 */
public class DynamicConstant {
	private final ReentrantReadWriteLock rwl = new ReentrantReadWriteLock();
	private final Lock rLock = rwl.readLock();
	private final Lock wLock = rwl.writeLock();
	/**
	 * 評価結果<br>
	 * 排他制御を行うので必ずgetter, setterを利用すること．
	 */
	private Object result;

	/** インスタンスを排他制御付きMapで管理する */
	static private Map<String, DynamicConstant> instances = new ConcurrentHashMap<String, DynamicConstant>();

	static public DynamicConstant create() {
		return create("");
	}
	/**
	 * 呼び出し元(クラス，メソッド，行)につき1つのインスタンスを生成します
	 * 
	 * @return
	 */
	static public DynamicConstant create(String initCode) {
		// throw可能なオブジェクトの情報に含まれるスタックトレースから情報を取得
		StackTraceElement ste = new Throwable().getStackTrace()[1];
		String callerIdentifier = ste.getClassName() + "#"
				+ ste.getMethodName() + ":"+ ste.getLineNumber();
		DynamicConstant instance = instances.get(callerIdentifier);
		if (instance == null) {
			instance = new DynamicConstant(callerIdentifier, initCode);
			instances.put(callerIdentifier, instance);
		}
		return instance;
	}

	private DynamicConstant(String callerIdentifier, String initCode) {
		ScriptEngineManager scManag = new ScriptEngineManager();
		final ScriptEngine engine = scManag.getEngineByName("JavaScript");

		final JFrame frame = new JFrame(callerIdentifier);
		frame.setSize(480, 100);
		frame.setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
		frame.setLayout(new FlowLayout(FlowLayout.RIGHT));
		
		frame.add(new JLabel("eval js: "));
		final JTextField scriptField = new JTextField(initCode, 36);
		JButton setButton = new JButton("Set");
		ActionListener setEvent = new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				try {
					engine.put("abc", 100);
					setResult(engine.eval(scriptField.getText()));
				} catch (ScriptException e1) {
					String message = e1.getMessage();
					int p = message.indexOf(':') + 2;
					message = message.substring(p);
					System.out.println(message.replaceAll("Unknown source", "Evaled code"));
				}
			}
		};
		scriptField.addActionListener(setEvent);
		setButton.addActionListener(setEvent);
		frame.add(scriptField);
		frame.add(setButton);
		
		frame.setVisible(true);
	}
	
	public Object getResult() {
		rLock.lock();
		try {
			return result;
		} finally {
			rLock.unlock();
		}
	}
	public String asString() {
		Object o = getResult();
		if (o instanceof String) return (String) o;
		if (o != null) {
			System.out.println("Cannnot cast to String from " + o.getClass().getName());
		}
		return "";
	}
	public double asDouble(double defaultValue) {
		Object o = getResult();
		if (o instanceof Double) return (Double) o;
		if (o != null) {
			System.out.println("Cannnot cast to numeric value from " + o.getClass().getName());
		}
		return defaultValue;
	}
	public double asDouble() {
		return asDouble(0.0);
	}
	public int asInteger(int defaultValue) {
		return (int) asDouble(defaultValue);
	}
	public int asInteger() {
		return (int) asDouble();
	}

	private void setResult(Object result) {
		wLock.lock();
		try {
			this.result = result;
		} finally {
			wLock.unlock();
		}
	}
	public boolean asBoolean(boolean defaultValue) {
		Object o = getResult();
		if (o instanceof Boolean) return (Boolean) o;
		if (o instanceof Double) return ((Double) o) != 0.0;
		if (o != null) {
			System.out.println("Cannnot cast to boolean value from " + o.getClass().getName());
		}
		return defaultValue;
	}
	public boolean asBoolean() {
		return asBoolean(false);
	}
	
	static public void main(String... args) {
		// ちょっと試してみたい人向け
		DynamicConstant inst = create("0.0");
		while(true) {
			System.out.println(inst.getResult());
			try {
				Thread.sleep(500);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}
}
