package rione.model;

import java.awt.Polygon;
import java.awt.geom.Point2D;
import java.util.*;

import rescuecore2.config.Config;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import rione.communication.CivilianVoiceListener;
import rione.model.ConfigConstants;
import rione.model.areapartition.AreaPartition;
import rione.model.areapartition.AreaPartitionBuilder;
import rione.model.flow.EnergyFlow;
import rione.model.route.pov.POVRouter;

/**
 * 災害空間の情報を保持するクラス
 */
public class AdvancedWorldModel extends StandardWorldModel {
	
	/** Configの定数群 */
	protected ConfigConstants config = null;
	public ConfigConstants getConfig() {
		return config;
	}

	protected POVRouter router = null;
	public POVRouter getRouter() {
		return router;
	}

	/** 建物の入り口 */
	private Entrance entrance = null;
	public Entrance getEntrance() {
		return entrance;
	}

	/** 時間を保持する */
	protected int time;
	public int getTime() {
		return time;
	}

	/** uniformを取り扱うオブジェクト */
	protected Uniform uniform = null;
	public Uniform getUniform() {
		return uniform;
	}

	/** 燃えている建物の集合 */
	protected BurningBuildings burningBuildings = null;
	public BurningBuildings getBurningBuildings() {
		return burningBuildings;
	}

	/** 閉塞の中にいるエージェントの集合 */
	protected LocateInBlockade locateInBlockade = null;
	public LocateInBlockade getLocateInBlockade() {
		return locateInBlockade;
	}

	/**
	 * Buildingの中に居るCivilianもしくはBuildingの中に居る埋まっているエージェントの集合
	 */
	protected BuriedHumans buriedHumans = null;
	public BuriedHumans getBuriedHumans() {
		return buriedHumans;
	}

	protected CriticalArea criticalArea = null;
	public CriticalArea getCriticalArea() {
		return criticalArea;
	}

	protected AreaPartition areaPartitions = null;
	public AreaPartition getAreaPartitions() {
		return areaPartitions;
	}

	protected EnergyFlow energyFlow = null;
	public EnergyFlow getEnergyFlow() {
		return energyFlow;
	}

	/** マップとして孤立しており、行くことができないEntityのリスト */
	protected IsolatedArea isolated = null;
	public IsolatedArea getIsolated() {
		return isolated;
	}

	protected TimeStamp timestamp = null;
	public TimeStamp getTimestamp() {
		return timestamp;
	}

	/** 視界内のEntity */
	private Set<EntityID> visibleEntity = null;

	/**
	 * WorldModelを初期化する
	 */
	public void initialize(StandardEntity agent, Config conf) {
		config = new ConfigConstants(conf, this);
		
		uniform = new Uniform(this);
		timestamp = new TimeStamp(this);
		burningBuildings = new BurningBuildings(this);
		buriedHumans = new BuriedHumans(this);
		if (agent instanceof Human) {
			isolated = new IsolatedArea(agent, this);
			entrance = new Entrance(this);
			router = new POVRouter((Human) agent, this);
		}
		criticalArea = new CriticalArea(this);
		locateInBlockade = new LocateInBlockade(this);
		areaPartitions = new AreaPartition(new AreaPartitionBuilder(this));
		energyFlow = new EnergyFlow(this);
	}

	/**
	 * WorldModelを更新
	 * 
	 * @param changed
	 */
	public void update(Human me, ChangeSet changed) {
		visibleEntity = changed.getChangedEntities();
		//見た場所にHumanが居なくても削除されないので手動で消す
	    for (StandardEntity se : getEntitiesOfType(AgentConstants.HUMANOIDS)) {
	    	Human hm = (Human) se;
		    if (me.getPosition(this).equals(hm.getPosition(this))
			    	&& !changed.getChangedEntities().contains(hm.getID())) {
				hm.undefinePosition();
		    }
		}
	    for (StandardEntity se : getEntitiesOfType(StandardEntityURN.BLOCKADE)) {
	    	Blockade block = (Blockade) se;
	    	if (me.getPosition().equals(block.getPosition())
			    	&& !changed.getChangedEntities().contains(block.getID())) {
	    		block.undefinePosition();
	    	}
	    }
		buriedHumans.update(changed);
		areaPartitions.update(this, changed);
		energyFlow.update(changed);
	}
	
	/**
	 * EnumSetからEntityの集合を得る
	 * 
	 * @param urns
	 *            URNのセット
	 * @return Entityの集合
	 */
	public Collection<StandardEntity> getEntitiesOfType(EnumSet<StandardEntityURN> urns) {
		Collection<StandardEntity> res = new HashSet<StandardEntity>();
		for (StandardEntityURN urn : urns) {
			res.addAll(getEntitiesOfType(urn));
		}
		return res;
	}

	/**
	 * X, Y座標から場所を得る
	 * 
	 * @param x
	 *            X座標
	 * @param y
	 *            Y座標
	 * @return 場所
	 */
	public Area getPositionFromCoordinates(int x, int y) {
		for (StandardEntity entity : getObjectsInRange(x, y, 1)) {
			if (entity instanceof Area) {
				Area area = (Area) entity;
				if (area.getShape().contains(x, y)) {
					return area;
				}
			}
		}
		return null;
	}

	/**
	 * 座標から場所を得る
	 * 
	 * @param coordinate
	 *            座標
	 * @return 場所
	 */
	public StandardEntity getPositionFromCoordinates(Pair<Integer, Integer> coordinate) {
		return getPositionFromCoordinates(coordinate.first(), coordinate.second());
	}

	/**
	 * AbstractWorldModeのメソッドをオーバーライドし 情報の更新時にタイムスタンプを更新するようにしている
	 */
	@Override
	public void merge(ChangeSet changeSet) {
		List<EntityID> newID = new ArrayList<EntityID>();
		for (EntityID id : changeSet.getChangedEntities()) {
			if (getEntity(id) == null) {
				newID.add(id);
			}
		}
		super.merge(changeSet);
		if (timestamp != null) {
			timestamp.merge(changeSet);
		}
		if (burningBuildings != null) {
			burningBuildings.update(this, changeSet);
		}
		if (locateInBlockade != null) {
			locateInBlockade.update(changeSet);
		}
		for (EntityID id : newID) {
			StandardEntity se = getEntity(id);
			if (se instanceof Civilian) {
				timestamp.addStateChangeListener(se);
			}
		}
	}

	/**
	 * 死亡予測時刻<br>
	 * ダメージがない場合はtimestep + 1<br>
	 * 判定できない場合は負<br>
	 * timestep以上に生きる場合もある<br>
	 * 使う場合は要注意<br>
	 * Damageの変化を考慮していないので，さらに早く死ぬ可能性がある
	 * 
	 * @param human
	 * @return
	 */
	public int getDeathTime(Human human) {
		if (human.isDamageDefined()) {
			if (human.getDamage() == 0) {
				return getConfig().timestep + 1;
			}
			else {
				int lastChangedTime = timestamp.getLastChangedTime(human.getID());
				if (lastChangedTime < 0) {
					lastChangedTime = time;
				}
				if (human.isHPDefined()) {
					return lastChangedTime + human.getHP() / human.getDamage() - 1;
				}
				else {
					return getConfig().timestep + 1;
				}
			}
		}
		return getConfig().timestep + 1;
	}

	public Set<EntityID> getVisibleEntitiesID() {
		return visibleEntity;
	}
	
	/**
	 * Entityが視界内にあるかどうかを返す
	 * @param id 判定するEntityのID
	 * @return 視界内にあるときtrue，ないときfalse
	 */
	public boolean isVisible(EntityID id) {
		return getVisibleEntitiesID().contains(id);
	}
	/**
	 * Entityが視界内にあるかどうかを返す
	 * @param id 判定するEntity
	 * @return 視界内にあるときtrue，ないときfalse
	 */
	protected boolean isVisible(StandardEntity entity) {
		return isVisible(entity.getID());
	}
	public CivilianVoiceListener createNewCivilianListener() {
		return new CivilianVoiceListener() {
			@Override
			public void hear(AKSpeak message) {
				final EntityID id = message.getAgentID();
				addNewCivilian(id);
			}
		};
	}
	
	public Civilian addNewCivilian(EntityID id) {
		StandardEntity existing = getEntity(id);
		if (existing == null) {
			// 初観測のCivilian
			Civilian civilian = new Civilian(id);
			timestamp.addStateChangeListener(civilian);
			addEntity(civilian);
			return civilian;
		}
		return null;
	}

	/** 
	 * 隣接しているBuilding Roadを返す
	 * 引数	対象のBuilding※StandardEntity型
	 * 			隣接の幅（メートル単位)
	 */
	public Set<StandardEntity> getNeighbours(StandardEntity e, int Ext){
		Set<StandardEntity> Neighbours = new HashSet<StandardEntity>();
		if (e instanceof Building){
			//隣接しているロード。//ついカっとなって追加した//11/19
			for(EntityID tmp2:((Building) e).getNeighbours()){
				Neighbours.add(getEntity(tmp2));
			}
			//隣接している建物。
			if(((Building) e).isEdgesDefined()){
				Ext*=1000;
				List<Edge> Edges = new ArrayList<Edge>();
				Edges=((Building) e).getEdges();
				Polygon ExtArea =new Polygon();
				Polygon baseArea = new Polygon();
				int n=1;
				Point2D.Double tmp1=new Point2D.Double();
				Point2D.Double tmp2=new Point2D.Double();
				Point2D.Double tmp3=new Point2D.Double();
				Point2D.Double tmp4=new Point2D.Double();
				Point2D.Double tmp5=new Point2D.Double();
				for(Edge Ee : Edges){baseArea.addPoint(Ee.getStartX(),Ee.getStartY());}
				for(Edge Ee : Edges)
				{
					for(Edge Ee1:Edges){
						if(Ee.getStart().equals(Ee1.getEnd())){
							tmp1.setLocation
							(
									Ee.getStartX()+(n*(Ee.getEndX()-Ee.getStartX()))/Math.hypot((double)Ee.getEndX()-(double)Ee.getStartX() , (double)Ee.getEndY()-(double)Ee.getStartY() )
									,
									Ee.getStartY()+(n*(Ee.getEndY()-Ee.getStartY()))/Math.hypot((double)Ee.getEndX()-(double)Ee.getStartX() , (double)Ee.getEndY()-(double)Ee.getStartY() )
							);
							tmp2.setLocation
							(
									Ee1.getEndX()+(n*(Ee1.getStartX()-Ee1.getEndX()))/Math.hypot((double)Ee1.getStartX()-(double)Ee1.getEndX() , (double)Ee1.getStartY()-(double)Ee1.getEndY() )
									,
									Ee1.getEndY()+(n*(Ee1.getStartY()-Ee1.getEndY()))/Math.hypot((double)Ee1.getStartX()-(double)Ee1.getEndX() , (double)Ee1.getStartY()-(double)Ee1.getEndY() )
							);

							tmp3.setLocation
							(
									(tmp1.x+tmp2.x)/2
									,
									(tmp1.y+tmp2.y)/2
							);
							tmp4.setLocation(Ee.getStartX(), Ee.getStartY());
							if(tmp3.x==tmp4.x&&tmp3.y==tmp4.y){continue;}
							else if(baseArea.contains(tmp3)){
								tmp5.setLocation(
										tmp4.x+(Ext*(tmp4.x-tmp3.x)/Math.hypot(tmp4.x-tmp3.x,tmp4.y-tmp3.y))
										,
										tmp4.y+(Ext*(tmp4.y-tmp3.y)/Math.hypot(tmp4.x-tmp3.x, tmp4.y-tmp3.y)));

							}
							else if (!baseArea.contains(tmp3)){
								tmp5.setLocation(
										tmp4.x+(Ext*(tmp3.x-tmp4.x)/Math.hypot(tmp3.x-tmp4.x,tmp3.y-tmp4.y))
										,
										tmp4.y+(Ext*(tmp3.y-tmp4.y)/Math.hypot(tmp3.x-tmp4.x, tmp3.y-tmp4.y)));
							}
							ExtArea.addPoint((int)tmp5.x, (int)tmp5.y);
						}
					}
				}
				for(StandardEntity checker:getObjectsInRange(e, 100*1000)){
					if(checker instanceof Building){
						if(((Building)checker).isEdgesDefined()){
							for(Edge Edger:((Building)checker).getEdges()){
								if(ExtArea.contains(Edger.getStartX(),Edger.getStartY())){
									Neighbours.add(checker);//中点も含めるようにしたい。
								}
								else	if(ExtArea.contains((Edger.getStartX()+Edger.getEndX())/2,(Edger.getStartY()+Edger.getEndY())/2)){
									Neighbours.add(checker);
								}
							}
						}
					}
				}
			}
		}
		else if (e instanceof Road){//11/19追加。
			for(EntityID Neighs:((Road) e).getNeighbours()){
				if(getEntity(Neighs) instanceof Road){Neighbours.add(getEntity(Neighs));}
				else if (getEntity(Neighs) instanceof Building){Neighbours.add(getEntity(Neighs));}
			}
		}
		return Neighbours;
	}

	public Area getNearestRoad(int x, int y) {
		Area result = getPositionFromCoordinates(x, y);
		if (result != null) return result;
		// 二分法で線形探索する範囲を絞る
		int l = 1;
		int r = (int) Math.hypot(
				Math.max(config.MAX_X, x) - Math.min(config.MIN_X, x),
				Math.max(config.MAX_Y, y) - Math.min(config.MIN_Y, y));
		// 1以上のサイズで最小のrange
		Collection<StandardEntity> minRange = null;
		while (r - l > 1000) {
			int m = (r + l) / 2;
			Collection<StandardEntity> range = getObjectsInRange(x, y, m);
			if (0 < range.size()) {
				r = m;
				if (minRange == null || minRange.size() < range.size()) {
					minRange = range;
				}
			}
			else {
				l = m;
			}
		}
		if (minRange == null) return null;
		double minDist = Double.MAX_VALUE;
		for (StandardEntity se : minRange) {
			if (se instanceof Area) {
				Area area = (Area) se;
				double dist = Math.hypot(area.getX() - x, area.getY() - y);
				if (dist < minDist) {
					minDist = dist;
					result = area;
				}
			}
		}
		return result;
	}

	public void setTime(int time) {
		this.time = time;
	}
}
