package rione.model;

/**
 * このクラスは複素数を表現します。複素数の計算を実行するメソッドを定義します。
 */
public class Complex {
    //インスタンス変数を持ちます。Complexオブジェクトが保持するxとyの2つの
    //double値はprivateなので、このクラスの外部からはアクセスできません。real()と
    //imag()を使ってアクセスします。
    private double x, y;
    
    /**
     * コンストラクタです。xとyを初期化します。
     * @param real
     * @param imag
     */
    public Complex(double real, double imag){
        this.x = real;
        this.y = imag;
    }
    
    /**
     * アクセサメソッドです。複素数の実部を返します。setReal()メソッドはないので、
     * 実部の設定はできません。つまりComplexクラスは不変クラスなのです。
     */
    public double real(){return x;}
    
    /**
     * 複素数の虚部を返すアクセサメソッドです。
     * @return
     */
    public double imag(){ return y;}
    
    /**
     * Complexを文字列に変換します。複素数が文字列に変換できるように、
     * Objectクラスのメソッドをオーバーライドしています。
     * System.out.println()などのメソッドで表示するときに便利です。
     */
    public String toString(){ return "{" + x + "," + y + "}"; }
    
    /**
     * 2つの複素数を取り、それらの和を返します
     * staticなので、このメソッドには「カレントインスタンス」や「this」オブジェクトは
     * ありません。次のように使います
     */
    public static Complex add(Complex a, Complex b){
        return new Complex(a.x + b.x, a.y + b.y);
    }
    
    /**
     * 2つの複素数を取り、それらの差を返します
     * @return
     */
    public static Complex sub(Complex a, Complex b){
        return new Complex(a.x - b.x, a.y - b.y);
    }
    
    /**
     * 2点間の長さを求める
     * @return
     */
    public static double abs(Complex a, Complex b){
        return (Math.sqrt((b.x - a.x) * (b.x - a.x) + (b.y-a.y) * (b.y-a.y)));
    }
    
    /**
     * 原点からの距離(複素数の絶対値)を求める
     * @return
     */
    public double abs(){
        return Math.sqrt(x * x + y * y);
    }
    
    /**
     * 原点からの距離(複素数の絶対値)を求める
     * @return
     */
    public double abs(double x, double y){
        return Math.sqrt(x * x + y * y);
    }    
    
    /**
     * 複素数の乗算を実行するstaticメソッドです。
     * @param a
     * @param b
     * @return
     */
    public static Complex mul(Complex a, Complex b){
        return new Complex(a.x * b.x - a.y * b.y, a.x * b.y + a.y * b.x);
    }
    
    /**
     * 複素数の乗算を実行するインスタンスメソッドです。
     * @param a
     * @param b
     * @return
     */    
    public Complex mul(double a){
    	return new Complex(x * a,y * a);
    }
    
    /**
     * 複素数の乗算を実行するインスタンスメソッドです。
     * @param a
     * @return
     */
    public Complex mul(Complex a){
        return new Complex(x * a.x - y * a.y, x * a.y + y * a.x);
    }
    
    
    public Complex div(Complex c) {
        double den = c.x*c.x + c.y*c.y;
        return new Complex((x*c.x + y*c.y)/den,(y*c.x - x*c.y)/den);
    }
    
    /**
     * 複素数の割り算を実行します
     * @param c
     * @param d
     * @return
     */    
    public Complex div(double d) {
        return new Complex(x / d,y / d);
    }
    
	/**
	 * 2点を円周上に持つ円の中心点
	 * 
	 * @param a
	 * @param b
	 * @return
	 */
	public static Complex getCenter(Complex a, Complex b, double radius ) {
		// ベクトルvを作成
		Complex v = Complex.sub(b, a);
		// 2点間の距離
		double d = v.abs();
		// 2点間の中心座標mを求める
		Complex m = Complex.add(a, (v.div(2.0)));
		// 線分(p1,p2)の単位法線ベクトル(線分(p1, p2)に対して垂直でかつ長さが1の線分)の作成
		Complex hv = new Complex(0, 1);
		Complex unv = Complex.mul(v, hv).div(d);
		// ベクトルxの作成
		double temp = Math.sqrt((radius*radius - d * d / 4.0));
		Complex x = unv.mul(temp);
		// 中心点を求める
		Complex center = Complex.add(m, x);
		//System.out.println("center "+center);
		return center;
	}
	
	//外積計算
	public static double cross(Complex a,Complex b){
		return a.x * b.y - a.y * b.x;
	}
	
	//a1,a2の線分とb1,b2の線分の交点を計算する
	public static Complex intersection(Complex a1,Complex a2,Complex b1,Complex b2){
		Complex b = sub(b2, b1);
		double d1 = cross(b,sub(a1, b1));
		if(d1 < 0){
			d1*=-1;
		}
		double d2 = cross(b,sub(a2, b1));
		if(d2 < 0){
			d2*=-1;
		}
		double t = d1 / (d1 + d2);
		Complex tc = new Complex(t, t);
		Complex temp = sub(a2,a1);
		return add(a1, mul(temp,tc));
	}
	
	

}