package rione.model;

import java.awt.Shape;
import java.util.Collection;
import java.util.HashSet;

import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import rione.communication.CommunicationUtil;
import rione.communication.MessageBitSection;
import rione.communication.Port;
import rione.io.BitArrayInputStream;

public class LocateInBlockade {
	
	private AdvancedWorldModel world;
	private HashSet<Human> blockedHumans = new HashSet<Human>();
	
	public LocateInBlockade(AdvancedWorldModel world) {
		this.world = world;
	}
	
	/**
	 * 閉塞内にHumanがいるかどうかを返す。
	 * 
	 * @param blockade
	 *            判定する閉塞
	 * @return 内部にHumanがいるときtrueいないときfalse
	 */
	public boolean isContainsHuman(Blockade blockade) {
		if (!blockade.isApexesDefined()) {
			return false;
		}
		for (StandardEntity entity : world.getEntitiesOfType(AgentConstants.HUMANOIDS)) {
			Pair<Integer, Integer> location = entity.getLocation(world);
			if (blockade.getShape().contains(location.first(), location.second())) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Humanが閉塞の内側にいるかどうかを返す。
	 * 
	 * @param human
	 *            判定するHuman
	 * @return HumanがいるBlockade
	 */
	public Blockade getStuck(Human human) {
		if (human.isXDefined() && human.isYDefined()) {
			return isLocateInBlockade(human.getLocation(world));
		}
		return null;
	}

	
	/**
	 * 座標が閉塞の内側かどうかを返す。
	 * 
	 * @param location
	 *            判定する座標
	 * @return 閉塞の内側ならtrue
	 */
	public Blockade isLocateInBlockade(Pair<Integer, Integer> location) {
		return isLocateInBlockade(location.first(), location.second());
	}
	/**
	 * 座標が閉塞の内側かどうかを返す。
	 * 
	 * @param x
	 *            判定するX座標
	 * @param y
	 *            判定するY座標
	 * @return 閉塞の内側ならtrue
	 */
	public Blockade isLocateInBlockade(int x, int y) {
		// TODO WorldModelから削除されていない実態のない閉塞で判定される可能性がある。
		for (StandardEntity se : world.getEntitiesOfType(StandardEntityURN.BLOCKADE)) {
			Blockade blockade = (Blockade) se;
			Shape s = blockade.getShape();
			// 計算時間的に不可
			//Shape s = ExpandApexes.expandApexes(blockade, 800);
			if (s != null && s.contains(x, y)) {
				return blockade;
			}
		}
		return null;
	}
	
	public void update(ChangeSet changeSet) {
		for (EntityID id : changeSet.getChangedEntities()) {
			StandardEntity entity = world.getEntity(id);
			switch (entity.getStandardURN()) {
			case CIVILIAN:
			case FIRE_BRIGADE:
			case AMBULANCE_TEAM:
			case POLICE_FORCE:
				final Human human = (Human) entity;
				if (getStuck(human) != null) {
					blockedHumans.add(human);
				}
				else {
					blockedHumans.remove(human);
				}
				break;
			default:
				break;
			}
		}
	}
	
	public Collection<Human> getLocateInBlocades() {
		return blockedHumans;
	}
	
	public Port createPort(final StandardEntity me, final CommunicationUtil comUtil) {
		return new Port() {
			private boolean added; 
			@Override
			public void init(ChangeSet changed) {
				added = false;
			}
			@Override
			public boolean hasNext() {
				return (me instanceof Human)
						&& getStuck((Human) me) != null
						&& !added;
			}
			@Override
			public MessageBitSection next() {
				// 中身なし
				added = true;
				return new MessageBitSection(20);
			}
			@Override
			public void read(EntityID sender, int time, BitArrayInputStream stream) {
				// 呼ばれたらsenderが詰まってる
				try {
					Human hm = (Human) world.getEntity(sender);
					if (hm == null) return;
					blockedHumans.add(hm);
				} catch(ClassCastException e) {
					// ノイズでCastに失敗した場合
				}
			}
		};
	}

}
