package rione.model;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import rescuecore2.misc.collections.LazyMap;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardPropertyURN;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.worldmodel.EntityListener;
import rescuecore2.worldmodel.Property;

final public class TimeStamp {
	/** 最後に状態が変化した時間 */
	private Map<EntityID, Integer> lastChangedTime = new HashMap<EntityID, Integer>();

	/** 最後に見た時間 */
	private Map<EntityID, Integer> lastSeenTime = new HashMap<EntityID, Integer>();
	
	/**
	 * 最後に情報を更新した時間
	 */
	private Map<EntityID, Map<StandardPropertyURN, Integer>> updatedPropertyTimeStamp = new LazyMap<EntityID, Map<StandardPropertyURN, Integer>>() {
		@Override
		public Map<StandardPropertyURN, Integer> createValue() {
			return new HashMap<StandardPropertyURN, Integer>();
		}
	};
	
	private AdvancedWorldModel world = null;
	
	public TimeStamp(AdvancedWorldModel w) {
		world = w;
		for (StandardEntity entity : world) {
			entity.addEntityListener(new StateChangeListener());
			if (entity instanceof Road) {
				entity.addEntityListener(new AreaBlockadeListener());
			}
		}
	}
	
	/**
	 * そのEntityを最後に見た時間を設定する
	 * 
	 * @param id
	 * @param timeStamp
	 */
	private void setLastSeenTime(EntityID id, int timeStamp) {
		lastSeenTime.put(id, timeStamp);
	}
	
	/**
	 * そのEntityを最後に見た時間<br>
	 * 見ていない場合は-1
	 * 
	 * @param id
	 * @return
	 */
	public int getLastSeenTime(EntityID id) {
		Integer time = lastSeenTime.get(id);
		return (time != null) ? time : -1;
	}
	
	/**
	 * Entityの状態が最後に変化した時間を得る
	 * 
	 * @param id
	 *            変化した時間を得たいEntity
	 * @return Entityの状態が最後に変化した時間
	 */
	public int getLastChangedTime(EntityID id) {
		Integer t = lastChangedTime.get(id);
		return t != null ? t.intValue() : -1;
	}
	
	/**
	 * 最後に状態が変化した時間を設定する．
	 * 通信によって外部から設定する時があるときのためにpublicにしてある．
	 * @param id
	 * @param time
	 */
	public void setLastChangedTime(EntityID id, int time) {
		lastChangedTime.put(id, time);
	}
	
	/**
	 * EntityのPropertyごとの最終更新情報を得る
	 * 
	 * @param id
	 * @return
	 */
	public Map<StandardPropertyURN, Integer> getPropertyTimeStampMap(EntityID id) {
		return updatedPropertyTimeStamp.get(id);
	}

	/**
	 * Propertyの最終更新時間を得る
	 * 
	 * @param id
	 * @param propertyUrn
	 * @return
	 */
	public int getPropertyTimeStamp(EntityID id, StandardPropertyURN propertyUrn) {
		Integer time = updatedPropertyTimeStamp.get(id).get(propertyUrn);
		return time != null ? time : -1;
	}

	/**
	 * Propertyの更新時間をセットする
	 * 
	 * @param id
	 * @param propertyUrn
	 * @param time
	 */
	private void setProtertyTimestamp(EntityID id, StandardPropertyURN propertyUrn, int time) {
		Map<StandardPropertyURN, Integer> map = updatedPropertyTimeStamp.get(id);
		map.put(propertyUrn, time);
	}
	
	/**
	 * @param id
	 * @return そのentityを見たことがあるかどうか
	 */
	public boolean haveSeen(EntityID id) {
		return this.getLastSeenTime(id) != -1;
	}

	/**
	 * EntityにStateChangeListenerを付加する 通信情報で初めて認識したEntityに使用する。
	 * 
	 * @param entity
	 *            StateChangeListenerを付加するEntity
	 */
	public void addStateChangeListener(Entity entity) {
		entity.addEntityListener(new StateChangeListener());
	}
	
	/**
	 * Entityの状態が変化したことを監視するListener
	 * 
	 * @author com
	 * 
	 */
	public class StateChangeListener implements EntityListener {
		@Override
		public void propertyChanged(Entity entity, Property property, Object oldValue, Object newValue) {
			final EntityID id = entity.getID();
			setLastChangedTime(id, world.getTime());
			try {
				// fromStringは内部でループを回しているので，速度が気になるようならMapにするとよい
				setProtertyTimestamp(id, StandardPropertyURN.fromString(property.getURN()), world.getTime());
			} catch(IllegalArgumentException e) {
				// propertyがStandardPropertyでないときにここに到達する
			}
		}
	}
	
	/**
	 * Area情報が更新された際、啓開済みの閉塞をWorldModelから削除するリスナー
	 * 
	 * @author com
	 * 
	 */
	public class AreaBlockadeListener implements EntityListener {

		@SuppressWarnings("unchecked")
		@Override
		public void propertyChanged(Entity entity, Property property, Object oldValue, Object newValue) {
			if (property.getURN().equals(StandardPropertyURN.BLOCKADES.toString())) {
				if (oldValue == null || newValue == null) {
					return;
				}
				List<EntityID> before = (List<EntityID>) oldValue;
				List<EntityID> after = (List<EntityID>) newValue;
				before.removeAll(after);
				for (EntityID id : before) {
					world.removeEntity(id);
				}
			}
		}
	}

	public void merge(ChangeSet changeSet) {
		for (EntityID e : changeSet.getChangedEntities()) {
			StandardEntity existingEntity = world.getEntity(e);
			Map<StandardPropertyURN, Integer> propertyTimestampMap = this.getPropertyTimeStampMap(existingEntity.getID());
			if (propertyTimestampMap == null) {
				propertyTimestampMap = new HashMap<StandardPropertyURN, Integer>();
			}
			for (Property p : changeSet.getChangedProperties(e)) {
				Property existingProperty = existingEntity.getProperty(p.getURN());
				existingProperty.takeValue(p);
				// Propertyのタイムスタンプを更新
				propertyTimestampMap.put(StandardPropertyURN.fromString(p.getURN()), world.getTime());
			}
			// Entityのタイムスタンプを更新
			this.setLastSeenTime(existingEntity.getID(), world.getTime());
		}
	}
}
