package rione.model.route.area;

import java.awt.Shape;
import java.util.*;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;
import rione.geom.ExpandApexes;
import rione.model.AdvancedWorldModel;
import rione.util.MyArea;

/**
 * 到達可能領域、および通行可能判定を行うクラス
 * 
 * @author com
 * 
 */
public class ReachableArea {

	// TODO configから値を取ってくる
	/** Platoonの斥力を含めたサイズ */
	private static final int PLATOON_SIZE = 800;

	/** WorldModelの参照 */
	private final AdvancedWorldModel world;

	private Set<Area> reachableAreas = new HashSet<Area>();
	private Set<EntityID> recognized = new HashSet<EntityID>();
	private Map<EntityID, Boolean> passableMap = new HashMap<EntityID, Boolean>();
	private Set<Road> passableIsFalse = new HashSet<Road>();

	public ReachableArea(AdvancedWorldModel world) {
		this.world = world;
		new MyArea();
	}

	/**
	 * Reachableに通信情報を反映させる
	 * 
	 * @param road
	 * @param passable
	 */
	public void update(Road road, boolean passable) {
		recognized.add(road.getID());
		passableMap.put(road.getID(), passable);
		if (!passable) {
			passableIsFalse.add(road);
		}
	}

	/**
	 * Area情報から、通行可能かどうかをを更新する。<br>
	 * 情報の正確性を確保するために、視界内の情報のに更新すること。
	 * 
	 * @param area
	 *            更新するArea
	 */
	private void update(Road road) {
		// まだ更新されていない情報または今(Agent内で)更新された情報ならば更新する
		// if (!recognized.contains(road.getID())) {
		updatePassable(road);
		recognized.add(road.getID());
		// }
		// updated = false;
	}

	private void updateReachable(StandardEntity me) {
		LinkedList<Area> llist = new LinkedList<Area>();
		reachableAreas.clear();
		Area myLocation = (Area) ((me instanceof Building) ? me : ((Human) me).getPosition(world));
		// 自分の位置から行ける範囲のAreaをreachableAreasにaddする
		llist.push(myLocation);
		do {
			Area area = llist.pop();
			reachableAreas.add(area);
			for (EntityID id : area.getNeighbours()) {
				Area neighbor = (Area) world.getEntity(id);
				if (reachableAreas.contains(neighbor)) {
					continue;
				}
				// passableMapで通れることがわかった場合、reachableAreasにadd
				Boolean passable = passableMap.get(id);
				if (neighbor instanceof Building || passable == null || passable) {
					llist.push(neighbor);
				}
			}
		} while (!llist.isEmpty());
	}

	public boolean isRecognized(Road road) {
		return recognized.contains(road.getID());
	}

	public ReachableType getReachableType(Area area) {
		if (recognized.contains(area.getID()) || area instanceof Building) {
			if (reachableAreas.contains(area)) {
				return ReachableType.True;
			}
			return ReachableType.False;
		}
		return ReachableType.False;
	}

	/**
	 * そのエリアが通行可能かどうかを返す。<br>
	 * 隣接するエリアとの境界線の中点がすべて同一の領域に含まれる場合のみ真を返す。
	 * 
	 * @param road
	 *            判定するRoad
	 * @return 通行可能かどうか
	 */
	public boolean isPassable(Road road) {
		Boolean passable = passableMap.get(road.getID());
		if (passable != null) {
			return passable;
		}
		return true;
	}

	/**
	 * Roadで from -> through -> to が通れるか返す
	 * 
	 * @param from
	 * @param through
	 * @param to
	 * @return
	 */
	public boolean isPassable(Road from, Road through, Road to) {
		// 間が通れるRoadの場合はtrue
		if (isPassable(through)) {
			return true;
		}
		// throughのRoadの形を取得
		MyArea area = new MyArea(through.getShape());
		// through内にBlockadeがある場合、全てのBlockadeを
		// throughのRoadの形(Area)から差し引く
		if (through.isBlockadesDefined()) {
			for (EntityID id : through.getBlockades()) {
				Blockade blockade = (Blockade) world.getEntity(id);
				Shape shape = expandApexes(blockade);
				if (shape != null) {
					area.subtract(new MyArea(shape));
				}
			}
		}
		// through内にBlockadeがなければtrue
		else {
			return true;
		}
		// Blockadeを差し引いたthroughが影も形もなければfalse
		if (area.isEmpty()) {
			return false;
		}
		// Blockadeを差し引いたthroughにfromとtoの形を足す
		area.add(new MyArea(from.getShape()));
		area.add(new MyArea(to.getShape()));
		// from -> through -> to が1つに繋がっている形ならばtrue
		// そうでなければfalse
		return area.isSingular();
	}

	/**
	 * 引数のRoadのPassableをUpdateする
	 * 
	 * @param road
	 */
	public void updatePassable(Road road) {
		// roadの形からBlockadeの形を差し引く
		MyArea area = new MyArea(road.getShape());
		if (road.isBlockadesDefined()) {
			for (EntityID id : road.getBlockades()) {
				Blockade blockade = (Blockade) world.getEntity(id);
				Shape shape = expandApexes(blockade);
				if (shape != null) {
					area.subtract(new MyArea(shape));
				}
			}
		}
		// roadが通れない状態ならばpassableをfalseにセットして返す
		if (area.isEmpty()) {
			passableMap.put(road.getID(), false);
			passableIsFalse.add(road);
			return;
		}
		// roadの全ての隣接Roadの形をroadの形に継ぎ足す
		for (EntityID id : road.getNeighbours()) {
			Area neighbor = (Area) world.getEntity(id);
			if (neighbor instanceof Road) {
				area.add(new MyArea(neighbor.getShape()));
			}
		}
		// passable <= roadが途切れていればfalse, 途切れていなければtrue
		boolean passable = area.isSingular();
		// passableをセットする
		passableMap.put(road.getID(), passable);
		if (!passable) {
			passableIsFalse.add(road);
		}
	}

	/**
	 * 対象のblockとそれに接するblockらが通路をふさいでいるかどうかの判定。
	 * 
	 * @param block
	 */
	public boolean isIgnorableBlockad(Blockade block) {
		return isIgnorableBlockad(block, true);
	}

	/**
	 * 対象のblockとそれに接するblockらが通路をふさいでいるかどうかの判定。
	 * 
	 * @param block
	 * @param includeBuilding
	 *            trueなら，Refuge以外のBuildingへの通路も塞いで居るか判定する
	 * @return
	 */
	public boolean isIgnorableBlockad(Blockade block, boolean includeBuilding) {
		if (world.getLocateInBlockade().isContainsHuman(block)) {
			return false;
		}

		if (!block.isPositionDefined()) {
			return false;
		}
		// 対象blockadeのAREA
		Area area = (Area) world.getEntity(block.getPosition());
		// ==対象blockadeを拡大しshape化
		MyArea newBlocks = new MyArea(expandApexes(block));
		// 頂点にかぶるかどうか
		List<Edge> edges = area.getEdges();
		for (Edge next : edges) {
			if (next.isPassable()) {
				int sx = next.getStartX();
				int sy = next.getStartY();
				int ex = next.getEndX();
				int ey = next.getEndY();
				if (newBlocks.contains((sx + ex) / 2, (sy + ey) / 2)) {
					return false;
				}
				// 閉塞の中心が他のエリアとの接続部分付近にあるときは無視しない
				if (getDistance((sx + ex) / 2, (sy + ey) / 2, block.getX(), block.getY()) < 4000) {
					return false;
				}
			}
		}

		for (EntityID id : area.getNeighbours()) {
			if (!includeBuilding && (world.getEntity(id) instanceof Building && !(world.getEntity(id) instanceof Refuge)))
				continue;
			Edge e = area.getEdgeTo(id);
			if (e.isPassable()) {
				int sx = e.getStartX();
				int sy = e.getStartY();
				int ex = e.getEndX();
				int ey = e.getEndY();
				if (newBlocks.contains((sx + ex) / 2, (sy + ey) / 2)) {
					return false;
				}
				if (getDistance(sx, sy, block.getX(), block.getY()) < 4000) {
					return false;
				}
				else if (getDistance(ex, ey, block.getX(), block.getY()) < 4000) {
					return false;
				}
			}
		}

		// if(b.contains(area.getX(),area.getY())){
		// return false;
		// }
		// ==対象blockadeのareaのshaep
		MyArea passableArea = new MyArea(area.getShape());
		// System.out.println(block);
		if (!(area instanceof Road))
			return false;
		// System.out.println("AREA:"+area);
		// ==対象エリアのblockadeのIDのリスト
		List<EntityID> blocksIDs = area.getBlockades();
		if (blocksIDs == null)
			return false;
		// 対象のblockadにくっつくblockadeくっつける
		for (EntityID blockID : blocksIDs) {
			MyArea tempNewBlocks = new MyArea(newBlocks);
			Blockade anotherBlock = (Blockade) world.getEntity(blockID);
			tempNewBlocks.add(new MyArea(anotherBlock.getShape()));
			if (tempNewBlocks.isSingular()) {
				// System.out.println("singlar");
				newBlocks.add(new MyArea(anotherBlock.getShape()));
			}
		}

		// neighberエリアをくっつけて通れるか確認
		java.util.List<EntityID> neighberID = area.getNeighbours();
		for (EntityID ID : neighberID) {
			Area neighboursArea = (Area) world.getEntity(ID);
			passableArea.add(new MyArea(neighboursArea.getShape()));
		}
		passableArea.subtract(newBlocks);

		if (passableArea.isEmpty()) {
			// System.out.println("empty");
			return false;
		}

		else if (passableArea.isSingular()) {
			// System.out.println("isSinglar");
			return true;
		}
		// System.out.println("notSinglar");
		return false;
	}

	private Shape expandApexes(Blockade blockade) {
		return ExpandApexes.expandApexes(blockade, PLATOON_SIZE);
	}

	public void update(StandardEntity me, Set<EntityID> visibleEntitiesID) {
		for (EntityID id : visibleEntitiesID) {
			StandardEntity entity = world.getEntity(id);
			if (entity instanceof Road) {
				update((Road)entity);
			}
		}
		updateReachable(me);
	}
	private static double getDistance(int x1, int y1, int x2, int y2) {
		return Math.hypot((x1 - x2), (y1 - y2));
	}
}
