package rione.model.route.pov;

import java.util.Collection;
import java.util.HashSet;
import java.util.PriorityQueue;
import java.util.Set;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rione.model.route.pov.graph.AreaNode;
import rione.model.route.pov.graph.EdgeNode;
import rione.model.route.pov.graph.PointNode;
import rione.model.route.pov.graph.PointOfVisivility;
import rione.model.route.pov.graph.RootNode;

public class POVSearch {
	
	final int routeThinkTime;
	final private PointOfVisivility pov;
	
	public POVSearch(final int thinkTime, PointOfVisivility pov) {
		routeThinkTime = thinkTime;
		this.pov = pov;
	}
	
	/**
	 * 足立法 
	 * 
	 * @author tatsuki
	 * @param end
	 * @param Human
	 * @param costFunc
	 * @param world
	 * @return
	 **/
public POVPath getAdachi(Area end,Area agents,CostFunction costFunc){
	return getAdachi(new RootNode(pov.get(end.getID()), end.getX(), end.getY()),agents,costFunc);
}


private POVPath getAdachi(AreaNode end,Area agents,CostFunction costFunc){
	
	final long timeLimit = routeThinkTime + System.currentTimeMillis();

	PriorityQueue<POVPath> queue = new PriorityQueue<POVPath>();
	Set<PointNode> visited = new HashSet<PointNode>();
	

			queue.offer(new POVPath(end,end.distance(agents.getX(), agents.getY())));

				do{
					POVPath currentRoute = queue.poll();//要素を取り出して削除
					if (visited.contains(currentRoute.getPoint())) continue; //一度訪れている場所は考慮しない
					visited.add(currentRoute.getPoint());
					if (currentRoute.getPoint() instanceof AreaNode){
						AreaNode searchArea = (AreaNode)currentRoute.getPoint();
						//自分のいる場所に到達したら終了
						if (searchArea.getBelong().equals(agents)) return currentRoute;
						for (EdgeNode next : searchArea.getNeighbours()) {
							if (visited.contains(next)) continue;
							// 次のArea（新しい探索先）をコストをつけてQueueに追加
							
							queue.offer(new POVPath(currentRoute, next,costFunc, end.distance(agents.getX(), agents.getY())));
						}
						// 計算時間の制限に引っ掛かりそうなときは生成途中のRouteを返す
						if (System.currentTimeMillis() > timeLimit) {
						//	System.err.println("Adachi-Route plan create time over."
							return currentRoute;
						}
					}
					else if(currentRoute.getPoint() instanceof EdgeNode){
						EdgeNode searchEdge = (EdgeNode)currentRoute.getPoint();
						for (AreaNode next : searchEdge.getNeighbours()) {
							if (visited.contains(next)) continue;
							// 次のArea（新しい探索先）をコストをつけてQueueに追加

							queue.offer(new POVPath(currentRoute, next, costFunc, end.distance(agents.getX(), agents.getY())));
						}
					}
				}while(!queue.isEmpty());

		return null;
}	
	/**
	 * A*探索 単一の目的地に移動する際に用いる
	 * 
	 * @param me
	 * @param destinations
	 * @param costFunc
	 * @param world
	 * @return
	 */
	public POVPath getAStarPath(Human me, Area destination, CostFunction costFunc) {
		return getAStarPath(new RootNode(pov.get(me.getPosition()), me.getX(), me.getY()),
				destination, costFunc);
	}
	/**
	 * A*探索 単一の目的地に移動する際に用いる
	 * 
	 * @param origin
	 * @param destinations
	 * @param costFunc
	 * @param world
	 * @return
	 */
	public POVPath getAStarPath(Area origin, Area destination, CostFunction costFunc) {
		return getAStarPath(new RootNode(pov.get(origin.getID()), origin.getX(), origin.getY()),
				destination, costFunc);
	}
	private POVPath getAStarPath(AreaNode from, Area destination, CostFunction costFunc) {
		if (destination == null) return null;
		final long timeLimit = routeThinkTime + System.currentTimeMillis();
		
		PriorityQueue<POVPath> queue = new PriorityQueue<POVPath>();
		Set<PointNode> visited = new HashSet<PointNode>();
		// オープンリストに追加
		queue.offer(new POVPath(from, from.distance(destination.getX(), destination.getY())));
		do {
			POVPath currentRoute = queue.poll(); // 要素を取り出して削除
			if (visited.contains(currentRoute.getPoint())) continue;
			visited.add(currentRoute.getPoint());
			if (currentRoute.getPoint() instanceof AreaNode) {
				AreaNode currentArea = (AreaNode) currentRoute.getPoint();
				if (currentArea.getBelong().equals(destination)) { // 目標地点に到達
					return currentRoute;
				}
				// 現在値からつながっている全てのAreaを調べる
				//System.out.println(currentArea.getNeighbours().size());
				for (EdgeNode next : currentArea.getNeighbours()) {
					if (visited.contains(next)) continue;
					/*System.out.print(currentArea);
					System.out.print(" -> ");
					System.out.println(next);*/
					// 次のArea（新しい探索先）をコストをつけてQueueに追加
					queue.offer(new POVPath(currentRoute, next, costFunc, currentArea.distance(destination.getX(), destination.getY())));
				}
				// 計算時間の制限に引っ掛かりそうなときは生成途中のRouteを返す
				if (System.currentTimeMillis() > timeLimit) {
					//(new AssertionError("A* POV-Route plan create time over.")).printStackTrace();
					System.err.println("A* POV-Route plan create time over.");
					return currentRoute;
				}
			}
			else if (currentRoute.getPoint() instanceof EdgeNode) {
				EdgeNode currentEdge = (EdgeNode) currentRoute.getPoint();
				for (AreaNode next : currentEdge.getNeighbours()) {
					if (visited.contains(next)) continue;
					// 次のArea（新しい探索先）をコストをつけてQueueに追加
					queue.offer(new POVPath(currentRoute, next, costFunc, currentEdge.distance(destination.getX(), destination.getY())));
				}
			}
		} while (!queue.isEmpty());
		return null;
	}
	
	public POVPath getDijkstraPath(Area origin, Set<StandardEntity> destinations, CostFunction costFunc) {
		if (destinations.size() == 0) return null;
		final long timeLimit = routeThinkTime + System.currentTimeMillis();
		
		PriorityQueue<POVPath> queue = new PriorityQueue<POVPath>();
		Set<PointNode> visited = new HashSet<PointNode>();
		// オープンリストに追加
		final AreaNode originNode = pov.get(origin.getID());
		queue.offer(new POVPath(originNode, 0.0));
		do {
			POVPath currentRoute = queue.poll(); // 要素を取り出して削除
			if (visited.contains(currentRoute.getPoint())) continue;
			visited.add(currentRoute.getPoint());
			if (currentRoute.getPoint() instanceof AreaNode) {
				AreaNode currentArea = (AreaNode) currentRoute.getPoint();
				if (destinations.contains(currentArea.getBelong())) { // 目標地点に到達
//					System.out.print("router remain: ");
//					System.out.println(timeLimit - System.currentTimeMillis());
					return currentRoute;
				}
				// 現在値からつながっている全てのAreaを調べる
				//System.out.println(currentArea.getNeighbours().size());
				for (EdgeNode next : currentArea.getNeighbours()) {
					if (visited.contains(next)) continue;
					//System.out.print(currentArea);
					//System.out.print(" -> ");
					//System.out.println(next);
					// 次のArea（新しい探索先）をコストをつけてQueueに追加
					queue.offer(new POVPath(currentRoute, next, costFunc, 0.0));
				}
				// 計算時間の制限に引っ掛かりそうなときは生成途中のRouteを返す
				if (System.currentTimeMillis() > timeLimit) {
					//(new AssertionError("Dijkstra POV-Route plan create time over.")).printStackTrace();
					System.err.println("Dijkstra POV-Route plan create time over.");
					return currentRoute;
				}
			}
			else if (currentRoute.getPoint() instanceof EdgeNode) {
				EdgeNode currentEdge = (EdgeNode) currentRoute.getPoint();
				for (AreaNode next : currentEdge.getNeighbours()) {
					// 次のArea（新しい探索先）をコストをつけてQueueに追加
					if (visited.contains(next)) continue;
					queue.offer(new POVPath(currentRoute, next, costFunc, 0.0));
				}
			}
		} while (!queue.isEmpty());
		return null;
	}
	
	/**
	 * 目的地が複数ある場合のA*<br>
	 * originに向かって探索．本当に早いかは謎．
	 * 返ってくるルートは他と逆向きになることに注意．
	 * @param origin
	 * @param destinations
	 * @param costFunc
	 * @return
	 */
	public POVPath getMultiAStarPath(Area origin, Collection<? extends StandardEntity> destinations, CostFunction costFunc) {
		if (origin == null || destinations == null) return null;
		final long timeLimit = routeThinkTime + System.currentTimeMillis();
		
		PriorityQueue<POVPath> queue = new PriorityQueue<POVPath>();
		Set<PointNode> visited = new HashSet<PointNode>();
		// オープンリストに追加
		for (StandardEntity se : destinations) {
			AreaNode start = pov.get(se.getID());
			queue.offer(new POVPath(start, start.distance(origin.getX(), origin.getY())));
		}
		do {
			POVPath currentRoute = queue.poll(); // 要素を取り出して削除
			if (visited.contains(currentRoute.getPoint())) continue;
			visited.add(currentRoute.getPoint());
			if (currentRoute.getPoint() instanceof AreaNode) {
				AreaNode currentArea = (AreaNode) currentRoute.getPoint();
				if (currentArea.getBelong().equals(origin)) { // 目標地点に到達
					return currentRoute;
				}
				// 現在値からつながっている全てのAreaを調べる
				//System.out.println(currentArea.getNeighbours().size());
				for (EdgeNode next : currentArea.getNeighbours()) {
					if (visited.contains(next)) continue;
					//System.out.print(currentArea);
					//System.out.print(" -> ");
					//System.out.println(next);
//					if (visited.contains(next))
//						continue;
					// 次のArea（新しい探索先）をコストをつけてQueueに追加
					queue.offer(new POVPath(currentRoute, next, costFunc, currentArea.distance(origin.getX(), origin.getY())));
				}
				// 計算時間の制限に引っ掛かりそうなときは生成途中のRouteを返す
				if (System.currentTimeMillis() > timeLimit) {
					//(new AssertionError("A* POV-Route plan create time over.")).printStackTrace();
					System.err.println("A* POV-Route plan (MultiDest) create time over.");
					return currentRoute;
				}
			}
			else if (currentRoute.getPoint() instanceof EdgeNode) {
				EdgeNode currentEdge = (EdgeNode) currentRoute.getPoint();
				for (AreaNode next : currentEdge.getNeighbours()) {
					if (visited.contains(next)) continue;
					// 次のArea（新しい探索先）をコストをつけてQueueに追加
					queue.offer(new POVPath(currentRoute, next, costFunc, currentEdge.distance(origin.getX(), origin.getY())));
				}
			}
		} while (!queue.isEmpty());
		return null;
	}
}
