package rione.model.route.pov.reachable;

import java.util.ArrayDeque;
import java.util.Collection;
import java.util.Deque;
import java.util.HashSet;
import java.util.Set;

import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;
import rione.model.AdvancedWorldModel;
import rione.model.AgentConstants;
import rione.model.route.pov.graph.AreaNode;
import rione.model.route.pov.graph.EdgeNode;
import rione.model.route.pov.graph.PassableDictionary;
import rione.model.route.pov.graph.PointNode;
import rione.model.route.pov.graph.PassableDictionary.PassableLevel;

/**
 * 再構築型のReachableArea<br>
 * 遅いがPASSABLEからUNPASSABLEの変化にも対応できる．
 * @author utisam
 *
 */
public class BFSReachableArea {
	
	private final Set<EntityID> reachableArea;
	private final Set<EntityID> sureReachableArea;
	private final int limit;
	
	public BFSReachableArea(AdvancedWorldModel world) {
		limit = world.getConfig().thinkTime / 2;
		Collection<StandardEntity> areas = world.getEntitiesOfType(AgentConstants.AREAS);
		reachableArea = new HashSet<EntityID>(areas.size());
		sureReachableArea = new HashSet<EntityID>(areas.size());
	}
	
	public void update(final AreaNode pos, final AdvancedWorldModel world) {
		updateSureReachable(pos, world);
		updateReachable(pos, world);
	}
	
	/**
	 * Passableのみで接続されたエリアを求めます．
	 * @param pos
	 * @param passableDic
	 */
	private void updateSureReachable(AreaNode pos, AdvancedWorldModel world) throws IllegalStateException {
		sureReachableArea.clear();
		Deque<PointNode> que = new ArrayDeque<PointNode>();
		Set<PointNode> visited = new HashSet<PointNode>();
		PassableDictionary passableDic = world.getRouter().getPassableDic();
		que.push(pos);
		//TODO LOGICAL_PASSABLEもUNPASSABLEと同等に考えて試す
		final long t = System.currentTimeMillis();
		do {
			PointNode currentNode = que.pop();
			visited.add(currentNode);
			if (currentNode instanceof AreaNode) {
				AreaNode areaNode = (AreaNode) currentNode;
				sureReachableArea.add(areaNode.getBelong().getID());
				for (EdgeNode edgeNode : areaNode.getNeighbours()) {
					if (visited.contains(edgeNode)) continue;
					if (passableDic.getPassableLevel(areaNode, edgeNode).ordinal()
							< PassableLevel.UNPASSABLE.ordinal()) {
						que.push(edgeNode);
					}
				}
			}
			else {
				EdgeNode edgeNode = (EdgeNode) currentNode;
				for (AreaNode areaNode : edgeNode.getNeighbours()) {
					if (visited.contains(areaNode)) continue;
					if (passableDic.getPassableLevel(areaNode, edgeNode).ordinal()
							< PassableLevel.UNPASSABLE.ordinal()) {
						que.push(areaNode);
					}
				}
			}
			if (System.currentTimeMillis() - t > limit) {
				throw new IllegalStateException();
			}
		} while (!que.isEmpty());
	}

	/**
	 * PassableとUnknownで繋がれたエリアを求めます．
	 * @param pos
	 * @param passableDic
	 */
	private void updateReachable(final AreaNode pos, AdvancedWorldModel world) throws IllegalStateException {
		reachableArea.clear();
		Deque<PointNode> que = new ArrayDeque<PointNode>();
		Set<PointNode> visited = new HashSet<PointNode>();
		que.push(pos);
		PassableDictionary passableDic = world.getRouter().getPassableDic();
		final long t = System.currentTimeMillis();
		do {
			PointNode currentNode = que.pop();
			visited.add(currentNode);
			if (currentNode instanceof AreaNode) {
				AreaNode areaNode = (AreaNode) currentNode;
				reachableArea.add(areaNode.getBelong().getID());
				for (EdgeNode edgeNode : areaNode.getNeighbours()) {
					if (visited.contains(edgeNode)) continue;
					if (passableDic.getPassableLevel(areaNode, edgeNode)
							!= PassableLevel.UNPASSABLE) {
						que.push(edgeNode);
					}
				}
			}
			else {
				EdgeNode edgeNode = (EdgeNode) currentNode;
				for (AreaNode areaNode : edgeNode.getNeighbours()) {
					if (visited.contains(areaNode)) continue;
					if (passableDic.getPassableLevel(areaNode, edgeNode)
							!= PassableLevel.UNPASSABLE) {
						que.push(areaNode);
					}
				}
			}
			if (System.currentTimeMillis() - t > limit) {
				throw new IllegalStateException();
			}
		} while (!que.isEmpty());
	}

	public boolean isSureReachable(EntityID id) {
		return sureReachableArea.contains(id);
	}
	
	public boolean isReachable(EntityID id) {
		return reachableArea.contains(id);
	}

}
