package rione.test;

import static org.junit.Assert.*;

import java.lang.reflect.Field;
import java.util.PriorityQueue;

import org.junit.Before;
import org.junit.Test;

import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rione.agent.Agent;
import rione.agent.CommunicationAgent;
import rione.agent.at.AmbulanceTeamAgent;
import rione.agent.fb.FireBrigadeAgent;
import rione.agent.pf.PoliceForceAgent;
import rione.communication.Port;
import rione.communication.CommunicationUtil;
import rione.communication.CommunicationManagerPartsBuilder;
import rione.communication.DummyPort;
import rione.communication.MessageBitSection;
import rione.io.BitArrayInputStream;
import rione.io.BitArrayOutputStream;
import rione.model.AgentConstants;


public class PortTest {

	static private final int dummyHeader = 100;
	static private final int dummyHeaderSize = 7;
	private MockWorldModel world;
	private CommunicationUtil comUtil;
	
	@Before
	public void setUp() throws Exception {
		world = MockWorldModel.getMockWorldModel();
		comUtil = new CommunicationUtil(world);
	}
	
	@Test
	public void testDummyPortListener() {
		// init
		final Port listener = new DummyPort();
		testPort(listener);
	}
	@Test
	public void testBuriedHumansPort() {
		final Port listener = world.getBuriedHumans().createAddPort(comUtil);
		testPort(listener);
	}
	@Test
	public void testAmbulanceTeamPort() {
		AmbulanceTeamAgent agent = new AmbulanceTeamAgent();
		testAgentPort(agent);
	}
	@Test
	public void testFireBrigadePort() {
		FireBrigadeAgent agent = new FireBrigadeAgent();
		testAgentPort(agent);
	}
	@Test
	public void testPoliceForcePort() {
		PoliceForceAgent agent = new PoliceForceAgent();
		testAgentPort(agent);
	}
	private void testAgentPort(CommunicationAgent<? extends Human> agent) {
		// worldがprivateだけど無理やり設定する
		try {
			Field f = null;
			f = Agent.class.getDeclaredField("world");
			f.setAccessible(true);
			f.set(agent, world);
		}
		catch (SecurityException e1) {
			e1.printStackTrace();
		}
		catch (NoSuchFieldException e1) {
			e1.printStackTrace();
		}
		catch (IllegalArgumentException e) {
			e.printStackTrace();
		}
		catch (IllegalAccessException e) {
			e.printStackTrace();
		}
		
		StandardEntity me = world.getEntity(agent.getID());
		CommunicationManagerPartsBuilder builder = new CommunicationManagerPartsBuilder(me, world);
		CommunicationUtil comUtil = builder.getCommunicationUtil();
		Port listener = agent.createPort(comUtil);
		testPort(listener);
	}
	
	@Test
	public void testCriticalAreaPort() {
		Port listener = world.getCriticalArea().createPort(comUtil);
		testPort(listener);
	}
	
	@Test
	public void testBurningBuildingsPort() {
		Port listener = world.getBurningBuildings().createAddPort(comUtil);
		testPort(listener);
	}

	@Test
	public void testAreaPartitionPort() {
		Port listener = world.getAreaPartitions().createPort(comUtil);
		testPort(listener);
	}
	
	@Test
	public void testLocateInBlockadePort() {
		for (StandardEntity se : world.getEntitiesOfType(AgentConstants.PLATOONS)) {
			Port listener = world.getLocateInBlockade().createPort(se, comUtil);
			testPort(listener);
		}
	}
	
	/**
	 * Portをテストする
	 * @param listener
	 */
	private static void testPort(Port listener) {
		listener.init(null);
		final PriorityQueue<MessageBitSection> que = createBitSectionsQueue(listener);
		int messageCount = que.size();
		final BitArrayInputStream in = new BitArrayInputStream(createBitArray(que));
		// InputStreamを全て読みだして0になったらok
		while (in.hasNext()) {
			assertEquals("Headerが変化しない", in.readBit(dummyHeaderSize), dummyHeader);
			listener.read(null, 0, in);
			messageCount--;
		}
		assertEquals("メッセージの数が変化しない", messageCount, 0);
	}
	
	private static byte[] createBitArray(final PriorityQueue<MessageBitSection> que) {
		final BitArrayOutputStream out = new BitArrayOutputStream();
		while (!que.isEmpty()) {
			final MessageBitSection sec = que.poll();
			// 書きこむ
			out.writeBit(sec.getHeaderNumber(), dummyHeaderSize);
			for (Pair<Integer, Integer> writePair : sec) {
				out.writeBit(writePair.first(), writePair.second());
			}
		}
		return out.getData();
	}
	
	private static PriorityQueue<MessageBitSection> createBitSectionsQueue(Port l) {
		final PriorityQueue<MessageBitSection> que = new PriorityQueue<MessageBitSection>();
		for (; l.hasNext();) {
			MessageBitSection sec = l.next();
			sec.setHeaderNumber(dummyHeader);
			que.add(sec);
		}
		return que;
	}
}
