package rione.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

/**
 * 集合内の要素が同じグループに属するかどうかを判定できるコレクションです．
 * @author utisam
 *
 * @param <T>
 */
public class UnionFindTree<T> implements Collection<T> {
	
	public final Map<T, T> parent;
	public final Map<T, Integer> rank;
	
	public UnionFindTree() {
		parent = new HashMap<T, T>();
		rank = new HashMap<T, Integer>();
	}
	
	public UnionFindTree(final int size) {
		parent = new HashMap<T, T>(size);
		rank = new HashMap<T, Integer>(size);
	}
	
	public UnionFindTree(final Collection<T> c) {
		parent = new HashMap<T, T>(c.size());
		rank = new HashMap<T, Integer>(c.size());
		addAll(c);
	}
	
	protected UnionFindTree(final Map<T, T> parentMap, final Map<T, Integer> rankMap) {
		parent = parentMap;
		rank = rankMap;
	}
	
	@Override
	public boolean add(final T o) {
		if (o == null) throw new NullPointerException();
		return parent.put(o, o) == null && rank.put(o, 0) == null;
	}
	
	@Override
	public boolean addAll(final Collection<? extends T> c) {
		boolean result = false;
		for (T x : c) {
			result = this.add(x) || result;
		}
		return result;
	}
	
	/**
	 * xの根を求める
	 * @param x
	 * @return xの根
	 */
	private T find(final T x) {
		if (x == null) return null;
		final T px = parent.get(x);
		if (px == null) return null;
		if (px.equals(x)) {
			return x;
		}
		final T y = find(px);
		parent.put(x, y);
		return y;
	}
	
	/**
	 * xとyの属する集合を併合する．
	 * 複数のSetを用いるより高速です．
	 * @param x
	 * @param y
	 */
	public void unite(final T x, final T y) {
		final T xRoot = find(x);
		final T yRoot = find(y);
		if (xRoot == null || xRoot.equals(yRoot)) return;
		final int xRank = rank.get(xRoot);
		final int yRank = rank.get(yRoot); 
		if (xRank < yRank) {
			parent.put(xRoot, yRoot);
		}
		else {
			parent.put(yRoot, xRoot);
			if (xRank == yRank) rank.put(xRoot, xRank + 1);
		}
	}
	
	/**
	 * xとyが同じ集合に属するかどうか．
	 * @param x
	 * @param y
	 * @return
	 */
	public boolean same(final T x, final T y) {
		final T xRoot = find(x);
		final T yRoot = find(y);
		return xRoot != null && yRoot != null && xRoot.equals(yRoot);
	}

	@Override
	public void clear() {
		parent.clear();
		rank.clear();
	}

	@Override
	public boolean contains(final Object o) {
		return parent.containsKey(o);
	}

	@Override
	public boolean containsAll(final Collection<?> c) {
		return parent.keySet().containsAll(c);
	}

	@Override
	public boolean isEmpty() {
		return parent.isEmpty();
	}

	@Override
	public Iterator<T> iterator() {
		return parent.keySet().iterator();
	}

	@Override
	public boolean remove(final Object o) {
		try {
			separate(o);
			return parent.remove(o) != null && rank.remove(o) != null;
		}
		catch (ClassCastException e) {
		}
		return false;
	}

	/**
	 * o を独立させます．
	 * 計算量はこのコレクションの保持する要素に対して O(n) 必要です．
	 * @param o
	 */
	public void separate(final Object o) {
		final T p = parent.get(o);
		if (o.equals(p)) {
			// oの親が自分自身なら，oを親にしている値も自分自身を親にしてunite
			final ArrayList<T> separated = new ArrayList<T>();
			for (Entry<T, T> entry : parent.entrySet()) {
				if (entry.getValue().equals(o)) {
					T key = entry.getKey();
					entry.setValue(key);
					separated.add(key);
				}
			}
			for (int i = 1; i < separated.size(); ++i) {
				unite(separated.get(0), separated.get(i));
			}
		}
		else {
			// oが親でないならoを親にしている値はoの親を親にする．
			for (Entry<T, T> entry : parent.entrySet()) {
				if (entry.getValue().equals(o)) {
					entry.setValue(p);
				}
			}
		}
	}

	@Override
	public boolean removeAll(final Collection<?> c) {
		boolean result = false;
		for (Object o : c) {
			result = remove(o) || result;
		}
		return result;
	}

	@Override
	public boolean retainAll(final Collection<?> c) {
		boolean result = false;
		for (Iterator<T> it = iterator(); it.hasNext();) {
			T next = it.next();
			if (!c.contains(next)) {
				it.remove();
				result = true;
			}
		}
		return result;
	}
	
	public void resetAll() {
		for (Entry<T, T> entry : parent.entrySet()) {
			entry.setValue(entry.getKey());
		}
		for (Entry<T, Integer> entry : rank.entrySet()) {
			entry.setValue(0);
		}
	}

	@Override
	public int size() {
		return parent.size();
	}

	@Override
	public Object[] toArray() {
		return parent.keySet().toArray();
	}

	@SuppressWarnings("hiding")
	@Override
	public <T> T[] toArray(T[] a) {
		return parent.keySet().toArray(a);
	}
	
}
